package monolith

  1. Overview
  2. Docs

Source file DelayedOutput.ml

1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
(******************************************************************************)
(*                                                                            *)
(*                                  Monolith                                  *)
(*                                                                            *)
(*                              François Pottier                              *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the GNU Lesser General Public License as published by the Free   *)
(*  Software Foundation, either version 3 of the License, or (at your         *)
(*  option) any later version, as described in the file LICENSE.              *)
(*                                                                            *)
(******************************************************************************)

let b =
  Buffer.create 1024

let () =
  GlobalState.register (fun () ->
    let snapshot = Buffer.contents b in
    fun () ->
      Buffer.clear b;
      Buffer.add_string b snapshot
  )

let[@inline] dprintf format =
  Printf.bprintf b format

let dump dst =
  Buffer.add_buffer dst b