Source file Engine.ml
1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
19
20
21
22
23
24
25
26
27
28
29
30
31
32
33
34
35
36
37
38
39
40
41
42
43
44
45
46
47
48
49
50
51
52
53
54
55
56
57
58
59
60
61
62
63
64
65
66
67
68
69
70
71
72
73
74
75
76
77
78
79
80
81
82
83
84
85
86
87
88
89
90
91
92
93
94
95
96
97
98
99
100
101
102
103
104
105
106
107
108
109
110
111
112
113
114
115
116
117
118
119
120
121
122
123
124
125
126
127
128
129
130
131
132
133
134
135
136
137
138
139
140
141
142
143
144
145
146
147
148
149
150
151
152
153
154
155
156
157
158
159
160
161
162
163
164
165
166
167
168
169
170
171
172
173
174
175
176
177
178
179
180
181
182
183
184
185
186
187
188
189
190
191
192
193
194
195
196
197
198
199
200
201
202
203
204
205
206
207
208
209
210
211
212
213
214
215
216
217
218
219
220
221
222
223
224
225
226
227
228
229
230
231
232
233
234
235
236
237
238
239
240
241
242
243
244
245
246
247
248
249
250
251
252
253
254
255
256
257
258
259
260
261
262
263
264
265
266
267
268
269
270
271
272
273
274
275
276
277
278
279
280
281
282
283
284
285
286
287
288
289
290
291
292
293
294
295
296
297
298
299
300
301
302
303
304
305
306
307
308
309
310
311
312
313
314
315
316
317
318
319
320
321
322
323
324
325
326
327
328
329
330
331
332
333
334
335
336
337
338
339
340
341
342
343
344
345
346
347
348
349
350
351
352
353
354
355
356
357
358
359
360
361
362
363
364
365
366
367
368
369
370
371
372
373
374
375
376
377
378
379
380
381
382
383
384
385
386
387
388
389
390
391
392
393
394
395
396
397
398
399
400
401
402
403
404
405
406
407
408
409
410
411
412
413
414
415
416
417
418
419
420
421
422
423
424
425
426
427
428
429
430
431
432
433
434
435
436
437
438
439
440
441
442
443
444
445
446
447
448
449
450
451
452
453
454
455
456
457
458
459
460
461
462
463
464
465
466
467
468
469
470
471
472
473
474
475
476
477
478
479
480
481
482
483
484
485
486
487
488
489
490
491
492
493
494
495
496
497
498
499
500
501
502
503
504
505
506
507
508
509
510
511
512
513
514
515
516
517
518
519
520
521
522
523
524
525
526
527
528
529
530
531
532
533
534
535
536
537
538
539
540
541
542
543
544
545
546
547
548
549
550
551
552
553
554
555
556
557
558
559
560
561
562
563
564
565
566
567
568
569
570
571
572
573
574
575
576
577
578
579
580
581
582
583
584
585
586
587
588
589
590
591
592
593
594
595
596
597
598
599
600
601
602
603
604
605
606
607
608
609
610
611
612
613
614
615
616
617
618
619
620
621
622
623
624
625
626
627
628
629
630
631
632
633
634
635
636
637
638
639
640
641
642
643
644
645
646
647
648
649
650
651
652
653
654
655
656
657
658
659
660
661
662
663
664
665
666
667
668
669
670
671
672
673
674
675
676
677
678
679
680
681
682
683
684
685
686
687
688
689
690
691
692
693
694
695
696
697
698
699
700
701
702
703
704
705
706
707
708
709
710
711
712
713
714
715
716
717
718
719
720
721
722
723
724
725
726
727
728
729
730
731
732
733
734
735
736
737
738
739
740
741
742
743
744
745
746
747
748
749
750
751
752
753
754
755
756
757
758
759
760
761
762
763
764
765
766
767
768
769
770
771
772
773
774
775
776
777
778
779
780
781
782
783
784
785
786
787
788
789
790
791
792
793
794
795
796
797
798
799
800
801
802
803
804
805
806
807
808
809
810
811
812
813
814
815
816
817
818
819
820
821
822
823
824
825
826
827
828
829
830
831
832
833
834
835
836
837
838
839
840
841
842
843
844
845
846
847
848
849
850
851
852
853
854
855
856
857
858
859
860
861
862
863
864
865
866
867
868
869
870
871
872
873
874
875
876
877
878
879
880
881
882
883
884
885
886
887
888
889
890
891
892
893
894
895
896
897
898
899
900
901
902
903
904
905
906
907
908
909
910
911
912
913
914
915
916
917
918
919
920
921
922
923
924
925
926
927
928
929
930
931
932
933
934
935
936
937
938
939
940
941
942
943
944
945
946
947
948
949
950
951
952
953
954
955
956
957
958
959
960
961
962
963
964
965
966
967
968
969
970
971
972
973
974
975
976
977
978
979
980
981
982
983
984
985
986
987
988
989
990
991
992
993
994
995
996
997
998
999
1000
1001
1002
1003
1004
1005
1006
1007
1008
1009
1010
1011
1012
1013
1014
1015
1016
1017
1018
1019
1020
1021
1022
1023
1024
1025
1026
1027
1028
1029
1030
1031
1032
1033
1034
1035
1036
1037
1038
1039
1040
1041
1042
1043
1044
1045
1046
1047
1048
1049
1050
1051
1052
1053
1054
1055
1056
1057
1058
1059
1060
1061
1062
1063
1064
1065
1066
1067
1068
1069
1070
1071
1072
1073
1074
1075
1076
1077
1078
1079
1080
1081
1082
1083
1084
1085
1086
1087
1088
1089
1090
1091
1092
1093
1094
1095
1096
1097
1098
1099
1100
1101
1102
1103
1104
1105
1106
1107
1108
1109
1110
1111
1112
1113
1114
1115
1116
1117
1118
1119
1120
1121
1122
1123
1124
1125
1126
1127
1128
1129
1130
1131
1132
1133
1134
1135
1136
1137
1138
1139
1140
1141
1142
1143
1144
1145
1146
1147
1148
1149
1150
1151
1152
1153
1154
1155
1156
1157
1158
1159
1160
1161
1162
1163
1164
1165
1166
1167
1168
1169
1170
1171
1172
1173
1174
1175
1176
1177
1178
1179
1180
1181
1182
1183
1184
1185
1186
1187
1188
1189
1190
1191
1192
1193
1194
1195
1196
1197
1198
1199
1200
1201
1202
1203
1204
1205
1206
1207
1208
1209
1210
1211
1212
1213
1214
1215
1216
1217
1218
1219
1220
1221
1222
1223
1224
1225
1226
1227
1228
1229
1230
1231
1232
1233
1234
1235
1236
1237
1238
1239
1240
1241
1242
1243
1244
1245
1246
1247
1248
1249
1250
1251
1252
1253
1254
1255
1256
1257
1258
1259
1260
1261
1262
1263
1264
1265
1266
1267
1268
1269
1270
1271
1272
1273
1274
1275
1276
1277
1278
1279
1280
1281
1282
1283
1284
1285
1286
1287
1288
1289
1290
1291
1292
1293
1294
1295
1296
1297
1298
1299
1300
1301
1302
1303
1304
1305
1306
1307
1308
1309
1310
1311
1312
1313
1314
1315
1316
1317
1318
1319
1320
1321
1322
1323
1324
1325
1326
1327
1328
1329
1330
1331
1332
1333
1334
1335
1336
1337
1338
1339
1340
1341
1342
1343
1344
1345
1346
1347
1348
1349
1350
1351
1352
1353
1354
1355
1356
1357
1358
1359
1360
1361
1362
1363
1364
1365
1366
1367
1368
1369
1370
1371
1372
1373
1374
1375
1376
1377
1378
1379
1380
1381
1382
1383
1384
1385
1386
1387
1388
1389
1390
1391
1392
1393
1394
1395
1396
1397
1398
1399
1400
1401
1402
1403
1404
1405
1406
1407
1408
1409
1410
1411
1412
1413
1414
1415
1416
1417
1418
1419
1420
1421
1422
1423
1424
1425
1426
1427
1428
1429
1430
1431
1432
1433
1434
1435
1436
1437
1438
1439
1440
1441
1442
1443
1444
1445
1446
1447
1448
1449
1450
1451
1452
1453
1454
1455
1456
1457
1458
1459
1460
1461
1462
1463
1464
1465
1466
1467
1468
1469
1470
1471
1472
1473
1474
1475
1476
1477
1478
1479
1480
1481
1482
1483
1484
1485
1486
1487
1488
1489
1490
1491
1492
1493
1494
1495
1496
1497
1498
1499
1500
1501
1502
1503
1504
1505
1506
1507
1508
1509
1510
1511
1512
1513
1514
1515
1516
1517
1518
1519
1520
1521
1522
1523
1524
1525
1526
1527
1528
1529
1530
1531
1532
1533
1534
1535
1536
1537
1538
1539
1540
1541
1542
1543
1544
1545
1546
1547
1548
1549
1550
1551
1552
1553
1554
1555
1556
1557
1558
1559
1560
1561
1562
1563
1564
1565
1566
1567
1568
1569
1570
1571
1572
1573
1574
1575
1576
1577
1578
1579
1580
1581
1582
1583
1584
1585
1586
1587
1588
1589
1590
1591
1592
1593
1594
1595
1596
1597
1598
1599
1600
1601
1602
1603
1604
1605
1606
1607
1608
1609
1610
1611
1612
1613
1614
1615
1616
1617
1618
1619
1620
1621
1622
1623
1624
1625
1626
1627
1628
1629
1630
open Printf
open Misc
open Error
open Eq
open Spec
open Env
open Ops
type document =
PPrint.document
include Debug.Make(struct let debug = false end)
let () =
Printexc.record_backtrace true
type arg =
| ArgConstant of document
| ArgVar of var
| ArgUnit
| ArgPair of arg * arg
| ArgOption of arg option
| ArgResult of (arg, arg) result
| ArgEither of (arg, arg) Either.t
| ArgList of arg list
| ArgAppConstant of document * arg
type expr =
| EOp of string
| EAppL of expr * arg
| EAppR of (document list -> document) * expr
type assertion =
document
type pat =
| PatWildcard
| PatBind of var
| PatRawVar of string
| PatObserveAgreement
| PatObserveDisagreement of assertion
| PatUnit
| PatPair of pat * pat
| PatOption of pat option
| PatResult of (pat, pat) result
| PatEither of (pat, pat) Either.t
| PatNil
| PatCons of pat
| PatAppR of (document list -> document) * pat
type instruction =
| I of pat * expr
type frozen_exn =
{ e: exn; backtrace: string }
let freeze e =
let backtrace = Printexc.get_backtrace() in
{ e; backtrace }
type failure =
| ReferenceFailure of frozen_exn
| CandidateFailure of frozen_exn
| ReferenceWrapperFailure of frozen_exn
| CandidateWrapperFailure of frozen_exn
| ObservationFailure
| CheckFailure of var * frozen_exn
type env =
value Env.env
(** The exception [PleaseBackOff] can be raised by the reference
implementation to indicate that this operation (or this particular choice
of arguments for this operation) should not be exercised. The reason could
be that it is not permitted, or that this has not been implemented yet.
This exception causes Monolith to back off and try something else. For this
reason, the reference implementation must not perform any side effect
before raising this exception. *)
exception PleaseBackOff
(** The exception [Unimplemented] can be raised by the candidate
implementation to indicate that this operation (or this particular choice
of arguments for this operation) should not be exercised. This exception
causes Monolith to abandon the current scenario and investigate other
scenarios. *)
exception Unimplemented
let break_instruction (i : instruction) : instruction list =
let waiting = Queue.create() in
let c = ref 0 in
let gensym () = let x = !c in c := x + 1; x in
let rec break_pat pat =
match pat with
| PatWildcard
| PatBind _
| PatObserveAgreement
| PatObserveDisagreement _
| PatUnit
| PatNil
| PatOption None
| PatRawVar _
-> pat
| PatPair (pat1, pat2) ->
PatPair (break_pat pat1, break_pat pat2)
| PatOption (Some pat) ->
PatOption (Some (break_pat pat))
| PatResult (Ok pat) ->
PatResult (Ok (break_pat pat))
| PatResult (Error pat) ->
PatResult (Error (break_pat pat))
| PatEither (Either.Left pat) ->
PatEither (Either.Left (break_pat pat))
| PatEither (Either.Right pat) ->
PatEither (Either.Right (break_pat pat))
| PatCons pat ->
PatCons (break_pat pat)
| PatAppR (constant, pat) ->
let x = sprintf "_t%d" (gensym()) in
let i = I (pat, EAppR (constant, EOp x)) in
Queue.add i waiting;
PatRawVar x
in
Queue.add i waiting;
let output = ref [] in
while not (Queue.is_empty waiting) do
let I (pat, expr) = Queue.take waiting in
output := I (break_pat pat, expr) :: !output
done;
List.rev !output
include struct
open Print
let print_var env x =
match Env.lookup env x with
| Value (SpecBaseAbstract (_, properties), _, _) ->
let base = properties.aty_var in
utf8format "%s%d" base (level x)
| Value _ ->
assert false
let ovar =
utf8format "observed"
let rec print_arg env arg =
match arg with
| ArgConstant constant ->
constant
| ArgVar x ->
print_var env x
| ArgUnit ->
unit
| ArgPair (arg1, arg2) ->
OCaml.tuple [ print_arg env arg1; print_arg env arg2 ]
| ArgOption oarg ->
option (print_arg env) oarg
| ArgResult rarg ->
result (print_arg env) (print_arg env) rarg
| ArgEither rarg ->
either (print_arg env) (print_arg env) rarg
| ArgList args ->
list (print_arg env) args
| ArgAppConstant (constant, arg) ->
parens (apply constant [ print_arg env arg ])
let rec print_expr env expr =
print_app env expr []
and print_app env expr (actuals : document list) =
match expr with
| EOp op ->
apply
(!^ op)
actuals
| EAppL (expr, arg) ->
print_app env expr (print_arg env arg :: actuals)
| EAppR (constant, expr) ->
constant
(parens (print_expr env expr) :: actuals)
let cons =
!^ " ::" ^^ break 1
let rec print_pat env pat =
match pat with
| PatWildcard ->
underscore
| PatBind x ->
print_var env x
| PatRawVar x ->
utf8string x
| PatObserveAgreement ->
underscore
| PatObserveDisagreement _ ->
ovar
| PatUnit ->
unit
| PatPair (pat1, pat2) ->
OCaml.tuple [ print_pat env pat1; print_pat env pat2 ]
| PatOption opat ->
option (print_pat env) opat
| PatResult rpat ->
result (print_pat env) (print_pat env) rpat
| PatEither rpat ->
either (print_pat env) (print_pat env) rpat
| PatNil
| PatCons _ ->
print_pat_list env [] pat
| PatAppR _ ->
assert false
and print_pat_list env accu pat =
match pat with
| PatCons (PatPair (pat1, pat2)) ->
print_pat_list env (pat1 :: accu) pat2
| PatCons _ ->
assert false
| PatNil ->
let pats = List.rev accu in
list (print_pat env) pats
| _ ->
assert (pat = PatWildcard);
let pats = List.rev (pat :: accu) in
group (lparen ^^ nest 2 (
flow_map cons (print_pat env) pats
) ^^ rparen)
let rec observations pat : assertion list =
match pat with
| PatWildcard
| PatBind _
| PatRawVar _
| PatObserveAgreement
| PatUnit
| PatOption None
| PatNil
-> []
| PatObserveDisagreement assertion ->
[assertion]
| PatPair (pat1, pat2) ->
observations pat1 @
observations pat2
| PatOption (Some pat)
| PatResult (Ok pat)
| PatResult (Error pat)
| PatEither (Either.Left pat)
| PatEither (Either.Right pat)
| PatCons pat
| PatAppR (_, pat)
-> observations pat
let observations (i : instruction) : assertion list =
let I (pat, _) = i in
let os = observations pat in
assert (List.length os <= 1);
os
let print_simple_instruction env (I (pat, expr)) =
toplevel_let (print_pat env pat) (print_expr env expr)
let print_instruction env i =
separate hardline (
List.map (print_simple_instruction env) (break_instruction i) @
observations i
)
let rec print_instructions env pc failure is =
match is with
| [] ->
empty
| i :: is ->
utf8format "(* @%02d *) " pc ^^ align (
print_instruction env i ^^
match is, failure with
| [], CheckFailure (x, _) ->
begin match Env.lookup env x with
| Value (SpecBaseAbstract (_, properties), rv, _) ->
let check = properties.aty_check rv in
hardline ^^
Code.print check [ print_var env x ]
^^ utf8format ";; (* this check fails! *)"
| _ ->
assert false
end
| _, _ ->
empty
) ^^
hardline ^^
let pc = pc + 1 in
print_instructions env pc failure is
let failure_type failure =
match failure with
| ReferenceFailure _ ->
"Failure in the reference implementation"
| CandidateFailure _ ->
"Failure in an instruction"
| CandidateWrapperFailure _ ->
sprintf "Failure in a wrapper on the candidate side"
| ReferenceWrapperFailure _ ->
sprintf "Failure in a wrapper on the reference side"
| ObservationFailure ->
"Failure in an observation: candidate and reference disagree"
| CheckFailure _ ->
"Failure in a well-formedness check"
let print_failure_exception pc failure : string =
match failure with
| ObservationFailure ->
""
| ReferenceFailure e
| CandidateFailure e
| ReferenceWrapperFailure e
| CandidateWrapperFailure e
| CheckFailure (_, e) ->
sprintf "(* @%02d: The exception %s was raised. *)\n\n%s"
pc
(Printexc.to_string e.e)
e.backtrace
let print_failure env past failure : string =
let b = Buffer.create 1024 in
let is = List.rev past in
let pc = List.length is in
bprintf b "(* @%02d: %s. *)\n" pc (failure_type failure);
DelayedOutput.dump b;
output b (print_instructions env 1 failure is);
Buffer.add_string b (print_failure_exception pc failure);
Buffer.contents b
let print_equality_assertion equal rv cv =
assert_ (Code.print equal [ ovar; rv ])
^^ !^ ";;"
^^ candidate_finds cv
end
exception Abort of string * int
let raise_abort env past failure =
let scenario = print_failure env past failure
and fuel = List.length past in
raise (Abort (scenario, fuel))
let no_requirement _ = true
exception ConstructArg of arg * failure
let rec construct_arg
: type r c . env -> (r, c) spec -> (r -> bool) -> arg * r * c
= fun env spec p ->
match spec with
| SpecBaseAbstract (tag, _) ->
Env.choose env (fun x (Value (spec', rv, cv)) ->
match spec' with
| SpecBaseAbstract (tag', _) ->
begin match Tag.equal tag tag' with
| Eq ->
let rv : r = rv
and cv : c = cv in
if p rv then Some (ArgVar x, rv, cv) else None
| exception Tag.RuntimeTagError ->
None
end
| _ ->
assert false
)
| SpecSubset (spec, q) ->
construct_arg env spec (fun v -> p v && q v)
| _ ->
let (_, rv, _) as outcome = construct_arg_no_requirement env spec in
if p rv then outcome else Gen.reject()
and construct_arg_no_requirement
: type r c . env -> (r, c) spec -> arg * r * c
= fun env spec ->
match spec with
| SpecBaseAbstract _ ->
construct_arg env spec no_requirement
| SpecSubset (spec, q) ->
construct_arg env spec q
| SpecConstructible { generate } ->
let code = generate() in
let arg = ArgConstant (Code.print code []) in
let v = Code.value code in
arg, v, v
| SpecUnit ->
ArgUnit, (), ()
| SpecPair (spec1, spec2) ->
let arg1, rv1, cv1 = construct_arg_no_requirement env spec1
and arg2, rv2, cv2 = construct_arg_no_requirement env spec2 in
ArgPair (arg1, arg2), (rv1, rv2), (cv1, cv2)
| SpecOption spec ->
if Gen.bool() then
ArgOption None, None, None
else
let arg, rv, cv = construct_arg_no_requirement env spec in
ArgOption (Some arg), Some rv, Some cv
| SpecResult (spec1, spec2) ->
if Gen.bool() then
let arg, rv, cv = construct_arg_no_requirement env spec1 in
ArgResult (Ok arg), Ok rv, Ok cv
else
let arg, rv, cv = construct_arg_no_requirement env spec2 in
ArgResult (Error arg), Error rv, Error cv
| SpecEither (spec1, spec2) ->
if Gen.bool() then
let arg, rv, cv = construct_arg_no_requirement env spec1 in
ArgEither (Either.Left arg), Either.Left rv, Either.Left cv
else
let arg, rv, cv = construct_arg_no_requirement env spec2 in
ArgEither (Either.Right arg), Either.Right rv, Either.Right cv
| SpecList (length, spec) ->
let n = length() in
let args, rvs, cvs = construct_arg_list env spec n [] [] [] in
ArgList args, rvs, cvs
| SpecMapOutof (rwrap, cwrap, spec) ->
begin
let arg, rv, cv = construct_arg_no_requirement env spec in
let arg = ArgAppConstant (Code.print cwrap [], arg) in
match rwrap rv with
| exception e ->
raise (ConstructArg (arg, ReferenceWrapperFailure (freeze e)))
| rv ->
match Code.value cwrap cv with
| exception e ->
raise (ConstructArg (arg, CandidateWrapperFailure (freeze e)))
| cv ->
arg, rv, cv
end
| SpecDeferred spec ->
construct_arg_no_requirement env (Lazy.force spec)
| SpecIfPol (spec, _) ->
construct_arg_no_requirement env spec
| SpecDeconstructible _ ->
assert false
| SpecTop ->
assert false
| SpecArrow _ ->
assert false
| SpecDependentArrow _ ->
assert false
| SpecNondet _ ->
assert false
| SpecMapInto _ ->
assert false
and construct_arg_list
: type r c . env -> (r, c) spec -> int ->
arg list -> r list -> c list -> arg list * r list * c list
= fun env spec n args rvs cvs ->
if n = 0 then
args, rvs, cvs
else
let arg, rv, cv = construct_arg_no_requirement env spec in
construct_arg_list env spec (n-1) (arg :: args) (rv :: rvs) (cv :: cvs)
let construct_arg_no_requirement env spec =
section @@ fun () ->
log "Constructing an argument.\n%!";
construct_arg_no_requirement env spec
let rec use
: type r c .
env -> (r, c) spec -> expr -> r -> c -> expr * (value, failure) result
= fun env spec expr rv cv ->
match spec with
| SpecArrow (spec1, spec2) ->
use env (SpecDependentArrow (spec1, fun _ -> spec2)) expr rv cv
| SpecDependentArrow (spec1, spec2) ->
log "use: dependent arrow.\n%!";
begin
match construct_arg_no_requirement env spec1 with
| exception ConstructArg (arg1, failure) ->
EAppL (expr, arg1), Error failure
| arg1, rv1, cv1 ->
let expr = EAppL (expr, arg1) in
let spec2 = spec2 rv1 in
match rv rv1 with
| exception e ->
expr, Error (ReferenceFailure (freeze e))
| rv2 ->
match cv cv1 with
| exception e ->
expr, Error (CandidateFailure (freeze e))
| cv2 ->
use env spec2 expr rv2 cv2
end
| SpecMapInto (rwrap, cwrap, spec) ->
log "use: transform (map_into).\n%!";
begin
let expr = EAppR (Code.print cwrap, expr) in
match rwrap rv with
| exception e ->
expr, Error (CandidateWrapperFailure (freeze e))
| rv ->
match Code.value cwrap cv with
| exception e ->
expr, Error (ReferenceWrapperFailure (freeze e))
| cv ->
use env spec expr rv cv
end
| SpecDeferred spec ->
log "use: deferred spec.\n%!";
use env (Lazy.force spec) expr rv cv
| SpecIfPol (_, spec) ->
log "use: ifpol.\n%!";
use env spec expr rv cv
| spec ->
log "use: done generating a context.\n%!";
expr, Ok (Value (spec, rv, cv))
let generate_instruction env : instruction * (value, failure) result =
log "Generating an instruction.\n%!";
section @@ fun () ->
let op, Value (spec, rv, cv) = pick() in
log "Picked operation \"%s\".\n%!" op;
let expr, result = use env spec (EOp op) rv cv in
I (PatWildcard, expr), result
exception Deconstruct of failure
let rec deconstruct : type r c .
env -> r -> c -> (r, c) spec -> bool * pat
= fun env rv cv spec ->
match spec, rv, cv with
| SpecTop, _, _ ->
log "deconstruct: top.\n%!";
true, PatWildcard
| SpecDeconstructible { equal; print }, _, _ ->
log "deconstruct: deconstructible type.\n%!";
if Code.value equal rv cv then
true, PatObserveAgreement
else
let assertion = print_equality_assertion equal (print rv) (print cv) in
false, PatObserveDisagreement assertion
| SpecBaseAbstract _, _, _ ->
log "deconstruct: abstract base type.\n%!";
let x = Env.limit env in
Env.bind env (Value (spec, rv, cv));
true, PatBind x
| SpecUnit, (), () ->
log "deconstruct: unit type.\n%!";
true, PatUnit
| SpecPair (spec1, spec2), (rv1, rv2), (cv1, cv2) ->
log "deconstruct: pair type.\n%!";
section @@ fun () ->
let ok1, pat1 = deconstruct env rv1 cv1 spec1 in
if ok1 then
let ok2, pat2 = deconstruct env rv2 cv2 spec2 in
if ok2 then
true, PatPair (pat1, pat2)
else
false, PatPair (PatWildcard, pat2)
else
false, PatPair (pat1, PatWildcard)
| SpecOption _, None, None ->
log "deconstruct: option type (None/None).\n%!";
true, PatOption None
| SpecOption spec, Some rv, Some cv ->
log "deconstruct: option type (Some/Some).\n%!";
let ok, pat = deconstruct env rv cv spec in
ok, PatOption (Some pat)
| SpecOption _, None, Some _ ->
log "deconstruct: option type (None/Some).\n%!";
false, PatOption None
| SpecOption _, Some _, None ->
log "deconstruct: option type (Some/None).\n%!";
false, PatOption (Some PatWildcard)
| SpecResult (spec, _), Ok rv, Ok cv ->
log "deconstruct: result type (Ok/Ok).\n%!";
let ok, pat = deconstruct env rv cv spec in
ok, PatResult (Ok pat)
| SpecResult (_, spec), Error rv, Error cv ->
log "deconstruct: result type (Error/Error).\n%!";
let ok, pat = deconstruct env rv cv spec in
ok, PatResult (Error pat)
| SpecResult _, Error _, Ok _ ->
log "deconstruct: option type (Error/Ok).\n%!";
false, PatResult (Error PatWildcard)
| SpecResult _, Ok _, Error _ ->
log "deconstruct: option type (Ok/Error).\n%!";
false, PatResult (Ok PatWildcard)
| SpecEither (spec, _), Either.Left rv, Either.Left cv ->
log "deconstruct: either type (Left/Left).\n%!";
let ok, pat = deconstruct env rv cv spec in
ok, PatEither (Either.Left pat)
| SpecEither (_, spec), Either.Right rv, Either.Right cv ->
log "deconstruct: either type (Right/Right).\n%!";
let ok, pat = deconstruct env rv cv spec in
ok, PatEither (Either.Right pat)
| SpecEither _, Either.Right _, Either.Left _ ->
log "deconstruct: option type (Right/Left).\n%!";
false, PatEither (Either.Right PatWildcard)
| SpecEither _, Either.Left _, Either.Right _ ->
log "deconstruct: option type (Left/Right).\n%!";
false, PatEither (Either.Left PatWildcard)
| SpecList _, [], [] ->
log "deconstruct: list type (Nil/Nil).\n%!";
true, PatNil
| SpecList (_, element), rv :: rvs, cv :: cvs ->
log "deconstruct: list type (Cons/Cons).\n%!";
let ok, pat =
deconstruct env (rv, rvs) (cv, cvs) (SpecPair (element, spec)) in
ok, PatCons pat
| SpecList _, [], _ :: _ ->
log "deconstruct: list type (Nil/Cons).\n%!";
false, PatNil
| SpecList _, _ :: _, [] ->
log "deconstruct: list type (Cons/Nil).\n%!";
false, PatCons (PatPair (PatWildcard, PatWildcard))
| SpecNondet spec, _, _ ->
log "deconstruct: SpecNondet.\n%!";
begin match rv cv with
| Valid rv ->
deconstruct env rv cv spec
| Invalid assertion ->
false, PatObserveDisagreement (assertion ovar)
| exception e ->
raise (Deconstruct (ReferenceFailure (freeze e)))
end
| SpecMapInto (rwrap, cwrap, spec), _, _ ->
log "deconstruct: transform (map_into).\n%!";
let ok, pat = deconstruct env (rwrap rv) (Code.value cwrap cv) spec in
ok, PatAppR (Code.print cwrap, pat)
| SpecDeferred spec, _, _ ->
deconstruct env rv cv (Lazy.force spec)
| SpecIfPol (_, spec), _, _ ->
deconstruct env rv cv spec
| SpecConstructible _, _, _ ->
assert false
| SpecSubset _, _, _ ->
assert false
| SpecArrow _, _, _ ->
assert false
| SpecDependentArrow _, _, _ ->
assert false
| SpecMapOutof _, _, _ ->
assert false
let perform_checks env past =
Env.foreach env (fun x (Value (spec, rv, cv)) ->
match spec with
| SpecBaseAbstract (_, properties) ->
let check = Code.value (properties.aty_check rv) in
begin try
check cv
with e ->
raise_abort env past (CheckFailure (x, (freeze e)))
end
| _ ->
assert false
)
let rec test fuel past env : unit =
if fuel > 0 then match generate_instruction env with
| exception Gen.OutOfInputData ->
let pc = List.length past in
printf "@%02d: Input data exhausted; end of this test.\n" pc
| exception Gen.Reject ->
log "Generation of an instruction failed; retrying.\n";
test fuel past env
| exception IllFormedSpec (op, msg) ->
error "in the specification of operation `%s`:\n%s" op msg
| exception e ->
printf "An exception was raised during generation!\n";
raise e
| _, Error (ReferenceFailure { e = PleaseBackOff; _ }) ->
test fuel past env
| _, Error (CandidateFailure { e = Unimplemented; _ }) ->
()
| i, Error failure ->
let past = i :: past in
raise_abort env past failure
| I (pat, expr), Ok v ->
assert (pat = PatWildcard);
let Value (spec, rv, cv) = v in
match deconstruct env rv cv spec with
| exception Deconstruct failure ->
let pat = PatWildcard in
let past = I (pat, expr) :: past in
raise_abort env past failure
| ok, pat ->
let past = I (pat, expr) :: past in
if not ok then begin
raise_abort env past ObservationFailure
end
else begin
let fuel = fuel - 1 in
perform_checks env past;
test fuel past env
end
let stored : env option ref =
ref None
let env fuel : env =
match !stored with
| None ->
let bound = 5 * fuel in
let dummy_value = Value (SpecUnit, (), ()) in
let env = Env.empty bound dummy_value in
stored := Some env;
env
| Some env ->
Env.clear env;
env
let test fuel =
log "Beginning one test run.\n%!";
section @@ fun () ->
let past = [] in
let env = env fuel in
test fuel past env;
log "This test run is finished.\n%!"
let run prologue fuel =
GlobalState.reset();
match prologue() with
| exception Gen.OutOfInputData ->
log "Prologue: input data exhausted; abandoning.\n"
| exception Gen.Reject ->
log "Prologue: generation failure; abandoning.\n"
| exception e ->
printf "Prologue: an exception was raised!\n";
raise e
| _ ->
test fuel
let init () =
DelayedOutput.dprintf " #require \"monolith\";;\n";
DelayedOutput.dprintf " module Sup = Monolith.Support;;\n";
GlobalState.save()
type prologue =
unit -> unit
type fuel =
int
type settings = {
source : string option;
(**An optional file name. If this file name is present, then the engine
runs in AFL mode and this file is used as a source of random bits.
If it is absent, then the engine runs in random testing mode and
/dev/urandom is used as a source of random bits. *)
timeout : float option;
(**An optional time limit, expressed in seconds. *)
max_scenarios : int;
(**In random testing mode, if [max_scenarios] failure scenarios are
found, then testing stops. *)
show_scenario : bool;
(**This Boolean flag determines whether failure scenarios should be
printed to the standard output channel. *)
save_scenario : bool;
(**This Boolean flag determines whether failure scenarios should be
saved to a file in the directory [./output/crashes]. *)
prologue : prologue;
(**A prologue. *)
fuel : fuel;
(**The desired initial amount of fuel. *)
}
let run_afl settings =
assert (settings.source <> None);
AflPersistent.run (fun () ->
Gen.with_source settings.source (fun () ->
try
run settings.prologue settings.fuel
with Abort (scenario, _fuel) ->
output_string stdout scenario;
flush stdout;
abort()
)
);
0
let[@inline] tick clock settings =
Clock.tick clock @@ fun () ->
printf "%s tests run so far (%s/s overall, %s/s now) (fuel = %d).\n%!"
(summarize (Clock.ticks clock))
(summarize (Clock.overall_ticks_per_second clock))
(summarize (Clock.current_ticks_per_second clock))
settings.fuel
let[@inline] show_and_save scenario settings =
if settings.show_scenario then begin
output_string stdout scenario;
print_newline();
flush stdout
end;
if settings.save_scenario then begin
let temp_dir = "./output/crashes" in
mkdirp temp_dir;
let prefix = sprintf "scenario.%03d." settings.fuel
and suffix = "" in
let _, oc = Filename.open_temp_file ~temp_dir prefix suffix in
output_string oc scenario;
close_out_noerr oc
end
type failures =
int
let rec run_random_loop settings clock (accu : failures) : failures =
try
while true do
run settings.prologue settings.fuel;
tick clock settings
done;
assert false
with
| Abort (scenario, fuel) ->
show_and_save scenario settings;
let settings = { settings with fuel }
and accu = accu + 1 in
if accu = settings.max_scenarios then
accu
else
run_random_loop settings clock accu
| Clock.Timeout ->
accu
let run_random settings : failures =
assert (settings.source = None);
Gen.with_source settings.source @@ fun () ->
let granularity = 1000 in
let timeout = settings.timeout in
let clock = Clock.make ?timeout granularity in
let accu = 0 in
run_random_loop settings clock accu
let parse prologue fuel : settings =
let source, timeout, fuel, max_scenarios, save_scenario, show_scenario =
ref None, ref None, ref fuel, ref max_int, ref true, ref true
in
let set_save_scenario b = save_scenario := b in
let set_show_scenario b = show_scenario := b in
let set_timeout t = timeout := Some (float_of_int t) in
let spec = Arg.align [
"--fuel",
Arg.Set_int fuel,
"<int> Set a fuel limit";
"--max-scenarios",
Arg.Set_int max_scenarios,
"<int> Stop if this number of failure scenarios is reached";
"--save-scenario",
Arg.Bool set_save_scenario,
"<bool> Enable/disable saving scenarios on disk";
"--show-scenario",
Arg.Bool set_show_scenario,
"<bool> Enable/disable showing scenarios on stdout";
"--timeout",
Arg.Int set_timeout,
"<int> Set a time limit (in seconds) (random testing mode only)";
] in
let usage = sprintf "Usage: %s <input file>" Sys.argv.(0) in
Arg.parse spec (fun s -> source := Some s) usage;
let source, timeout, fuel, max_scenarios, save_scenario, show_scenario =
!source, !timeout, !fuel, !max_scenarios, !save_scenario, !show_scenario
in
{ source; timeout;
max_scenarios; save_scenario; show_scenario;
prologue; fuel }
let run settings =
init();
let failures =
match settings.source with
| Some _ ->
run_afl settings
| None ->
run_random settings
in
exit (if failures > 0 then 1 else 0)
let main ?(prologue = ignore) fuel =
let settings = parse prologue fuel in
run settings