package monolith

  1. Overview
  2. Docs

Source file Eq.ml

1
2
3
4
5
6
7
8
9
10
11
12
13
14
15
16
17
18
(******************************************************************************)
(*                                                                            *)
(*                                  Monolith                                  *)
(*                                                                            *)
(*                              François Pottier                              *)
(*                                                                            *)
(*  Copyright Inria. All rights reserved. This file is distributed under the  *)
(*  terms of the GNU Lesser General Public License as published by the Free   *)
(*  Software Foundation, either version 3 of the License, or (at your         *)
(*  option) any later version, as described in the file LICENSE.              *)
(*                                                                            *)
(******************************************************************************)

(* The equality GADT. A value of type [('a, 'b) eq] is a runtime witness of
   the equality of the types ['a] and ['b]. *)

type ('a, 'b) eq =
  | Eq : ('a, 'a) eq